#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Weight quantizers API"""

# -*- coding: utf-8 -*-
from __future__ import absolute_import
import tensorflow.keras.backend as K
import tensorflow as tf
from tensorflow.keras.layers import Layer


class WeightQuantizer(Layer):
    """The base class for all weight quantizers.

    This base class must be overloaded as well as the two functions `quantize`
    and `scale_factor`.

    Quantizers derived from this class must be symmetric uniform mid-tread
    quantizers, in order to be compatible with the conversion into an Akida
    model. Quantization is usually done in two steps:

    #. The weights must be first quantized on integer values in the
       range imposed by the bitwidth, e.g. from -7 to 7 for a 4-bit
       quantization.
    #. These integer weights are then reconstructed to float discretized
       values, in the range of the original weights. For example, 4-bit integer
       weights are reconstructed on a grid from -7*qstep to 7*qstep, where
       qstep is the quantization step size between two levels of the uniform
       grid.

    For a full explanation about mid-tread uniform quantization, one can take a
    look at `the Wikipedia page <https://en.wikipedia.org/wiki/Quantization_(signal_processing)#Example>`_.

    The `quantize` function takes as inputs the original weights and must
    return the reconstructed float values after quantization. The
    `scale_factor` function must return the factor used to transform the float
    reconstructed weights into the integer values obtained after step 1. In other
    words, given a set of float weights "w":

     quantize(w) * scale_factor(w) is a set of integer weights.

    The bitwidth defines the number of quantization levels on which the
    weights will be quantized. For instance, a 4-bit quantization gives
    integer values between -7 and 7. More generally, for a n-bit
    quantization, values are ranged from -kmax to kmax where kmax is
    (2^(n-1) - 1).

    Args:
        bitwidth (int): the quantization bitwidth.
    """

    def __init__(self, bitwidth, **kwargs):
        self.bitwidth_ = int(bitwidth)
        self.kmax_ = (2.**(bitwidth - 1) - 1)
        super().__init__(**kwargs)

    def quantize(self, w):
        """Quantizes the specified weights Tensor.

        This function must return a tf.Tensor containing float weights
        discretized on a uniform grid based on the scale factor "sf".
        In other words, the discretized weights must be values among:
        -kmax/sf, ..., -2/sf, -1/sf, 0, 1/sf, 2/sf, ..., kmax/sf

        Args:
            w (:obj:`tensorflow.Tensor`): the weights Tensor to quantize.

        Returns:
            :obj:`tensorflow.Tensor`: a Tensor of quantized weights.

        """
        raise NotImplementedError()

    def scale_factor(self, w):
        """Evaluates the scale factor for the specified weights tf.Tensor.

        This function returns the scale factor to get the quantized integer
        weights from the reconstructed float weights. It is equal to the
        inverse of the quantization step size.

        The scale factor can be a scalar that is applied on the whole tensor
        of weights. It can also be a vector of length the number of filters,
        where each value applies to the weights of the corresponding output
        filter. This is called a per-axis quantization, as opposed to a
        per-tensor quantization. The number of filters is usually the last
        dimension of the weights tensor. More details are given
        `here <https://www.tensorflow.org/lite/performance/quantization_spec?hl=sv#per-axis_vs_per-tensor>`__

        Note that the `quantizer_dw` of a depthwise convolution in a
        QuantizedSeparableConv2D layer must imperatively return a scalar scale
        factor.

        Args:
          w (:obj:`tensorflow.Tensor`): the weights Tensor to quantize.

        Returns:
          :obj:`tensorflow.Tensor`: a Tensor containing a single scalar value.

        """
        raise NotImplementedError()

    @property
    def bitwidth(self):
        """Returns the bitwidth of the quantizer"""
        return self.bitwidth_

    def get_config(self):
        config = super().get_config()
        config.update({'bitwidth': self.bitwidth_})
        return config


class LinearWeightQuantizer(WeightQuantizer):
    """An abstract linear weight quantizer

    This abstract class proposes a linear symmetric and uniform quantization
    function. The "linear" term here means that there is no non-linear
    transformation of the weights before the uniform quantization.

    The `scale_factor` function must be overloaded.
    """

    def scale_factor(self, w):
        raise NotImplementedError()

    def quantize(self, w):
        """Linearly quantizes the input weights on a symmetric uniform grid
        based on the scale factor.

        The input weights are directly rounded to the closest discretized
        value, without any transformation on the input weights.

        The gradient is estimated using the Straight-Through Estimator (STE),
        i.e. the gradient is computed as if there were no quantization.
        """
        sf = self.scale_factor(w)
        return tf.clip_by_value(round_through(w * sf), -self.kmax_,
                                self.kmax_) / sf


class StdWeightQuantizer(LinearWeightQuantizer):
    """A uniform quantizer.

    Quantizes the specified weights into 2^bitwidth-1 values centered on zero.
    E.g. with bitwidth = 4, 15 quantization levels: from -7 * qstep to 7 * qstep
    with qstep being the quantization step. The quantization step is defined by:

     qstep = threshold * std(W) / max_value

    with max_value being 2^(bitwidth-1) - 1. E.g with bitwidth = 4, max_value = 7.

    All values below or above threshold * std(W) are automatically assigned to
    the min (resp max) value.

    Args:
        threshold (int): the standard deviation multiplier used to exclude
            outliers.
        bitwidth (int): the quantizer bitwidth defining the number of
            quantized values.
    """

    def __init__(self, threshold=3, bitwidth=4, **kwargs):
        # Having a cast guarantees a check when the parameters are not numbers
        # (e.g.: None)
        self.threshold_ = float(threshold)
        # Initialize parent to store the bitwidth
        super().__init__(bitwidth, **kwargs)

    def sigma_scaled_(self, w):
        """Get the range on which the weights are quantized. This quantizer
        discretizes weights in the range:

         [-threshold * std(weights) ; threshold * std(weights)]
        """
        return K.std(w) * self.threshold_

    def scale_factor(self, w):
        return self.kmax_ / self.sigma_scaled_(w)

    def get_config(self):
        config = super().get_config()
        config.update({'threshold': self.threshold_})
        return config


class TrainableStdWeightQuantizer(LinearWeightQuantizer):
    """A trainable weight quantizer.

    Quantizes the specified weights into 2^bitwidth-1 values centered on zero.
    E.g. with bitwidth = 4, 15 quantization levels: from -7 * qstep to 7 * qstep
    with qstep being the quantization step. The quantization step is defined by:

     qstep = threshold * std(W) / max_value

    with:

     - max_value being 2^(bitwidth-1) - 1. E.g with bitwidth = 4, max_value = 7.
     - threshold a trainable parameter whose initial value can be specified.

    All values below or above threshold * std(W) are automatically assigned to
    the min (resp max) value.

    This is the trainable version of the StdWeightQuantizer class.

    Args:
        threshold (int): the initial value of the standard deviation
            multiplier used to exclude outliers.
        bitwidth (int): the quantizer bitwidth defining the number of
            quantized values.

    """

    def __init__(self, threshold=3, bitwidth=4, **kwargs):
        super().__init__(bitwidth, **kwargs)
        self.threshold_initializer_ = threshold
        self.threshold_ = tf.Variable(name=f"{self.name}/threshold",
                                      initial_value=float(
                                          self.threshold_initializer_),
                                      dtype=tf.float32)

    def sigma_scaled_(self, w):
        """Get the range on which the weights are quantized. This quantizer
        discretizes weights in the range:

         [-threshold * std(weights) ; threshold * std(weights)]
        """
        return tf.math.reduce_std(w) * self.threshold_

    def scale_factor(self, w):
        return self.kmax_ / self.sigma_scaled_(w)

    def get_config(self):
        config = super().get_config()
        config.update({'threshold': self.threshold_initializer_})
        return config


class MaxQuantizer(LinearWeightQuantizer):
    """A quantizer that relies on maximum range.

    Quantizes the specified weights into 2^bitwidth-1 values centered on zero.
    E.g. with bitwidth = 4, 15 quantization levels: from -7 * qstep to 7 * qstep
    with qstep being the quantization step. The quantization step is defined by:

     qstep = max_range / max_value

    with:

     - max_range = max(abs(W))
     - max_value = 2^(bitwidth-1) - 1. E.g with bitwidth = 4, max_value = 7.

    Args:
        bitwidth (int): the quantizer bitwidth defining the number of
            quantized values.

    """

    def __init__(self, bitwidth=4, **kwargs):
        # Initialize parent to store the bitwidth
        super().__init__(bitwidth, **kwargs)

    @staticmethod
    def max_range_(w):
        """Get the range on which the weights are quantized. This quantizer
        discretizes weights in the range:

         [-max(weights) ; max(weights)]
        """
        return tf.math.reduce_max(tf.math.abs(w))

    def scale_factor(self, w):
        return self.kmax_ / self.max_range_(w)


class MaxPerAxisQuantizer(MaxQuantizer):
    """A quantizer that relies on maximum range per axis.

    Quantizes the specified weights into 2^bitwidth-1 values centered on zero.
    E.g. with bitwidth = 4, 15 quantization levels: from -7 * qstep to 7 * qstep
    with qstep being the quantization step. The quantization step is defined by:

     qstep = max_range / max_value

    with:

     - max_range = max(abs(W))
     - max_value = 2^(bitwidth-1) - 1. E.g with bitwidth = 4, max_value = 7.

    This is an evolution of the MaxQuantizer that defines the max_range per
    axis.

    The last dimension is used as axis, meaning that the scaling factor is a
    vector with as many values as "filters", or "neurons".

    Note: for a DepthwiseConv2D layer that has a single filter, this
    quantizer is strictly equivalent to the MaxQuantizer.

    """

    def max_range_(self, w):
        red_range = tf.range(tf.rank(w) - 1)
        return tf.math.reduce_max(tf.math.abs(w), red_range)


class ScaleFactorQuantizer(LinearWeightQuantizer):
    """This weight quantizer uses fixed scale factors given at instantiation
    to compute quantized weights.

    Unlike other quantizers where scale factors
    are calculated based on the weights to quantize, the scale factors here
    are fixed. It is not meant to be used directly but only to "freeze" the
    quantization range of another quantizer. For example, it is used to
    freeze any quantizer before folding BatchNormalization layers.
    Manipulating layers and weights with this quantizer must be done with care,
    as it might require to update the frozen scale factors.

    Args:
        bitwidth (int): the quantization bitwidth.
        scale_factors(float or list): the scale factors for all filters. If the
            scale factor is identical for all filters, a scalar can be given.

    """

    def __init__(self, bitwidth, scale_factors, **kwargs):
        self.scale_factors = tf.constant(scale_factors)
        super().__init__(bitwidth, **kwargs)

    def scale_factor(self, w):
        return self.scale_factors

    def get_config(self):
        config = super().get_config()
        config.update({'scale_factors': self.scale_factors.numpy()})
        return config


def get(identifier):
    """Returns the weight quantizer corresponding to the identifier.

    The 'identifier' input can take two types: either a weight quantizer
    instance, or a dictionary corresponding to the config serialization
    of a weight quantizer.

    Args:
        identifier (WeightQuantizer or dict): either a WeightQuantizer
            instance or a configuration dictionary to deserialize.

    Returns:
        :obj:`cnn2snn.WeightQuantizer`: a weight quantizer
    """
    if isinstance(identifier, WeightQuantizer):
        return identifier
    if isinstance(identifier, dict):
        return tf.keras.utils.deserialize_keras_object(
            identifier,
            custom_objects={
                'StdWeightQuantizer': StdWeightQuantizer,
                'TrainableStdWeightQuantizer': TrainableStdWeightQuantizer,
                'MaxQuantizer': MaxQuantizer,
                'MaxPerAxisQuantizer': MaxPerAxisQuantizer,
                'ScaleFactorQuantizer': ScaleFactorQuantizer
            })

    raise ValueError(f"Could not interpret identifier {identifier} "
                     "for a weight quantizer object")


def round_through(x):
    """Element-wise rounding to the closest integer with full gradient propagation.
    A trick from [Sergey Ioffe](http://stackoverflow.com/a/36480182).

    """
    rounded = K.round(x)
    return x + K.stop_gradient(rounded - x)


def ceil_through(x):
    """Element-wise ceiling operation (to the closest greater integer) with
    full gradient propagation.

    """
    ceiling_value = tf.math.ceil(x)
    return x + K.stop_gradient(ceiling_value - x)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Parsing functions that are able to generate an Akida model from a keras model.

"""
import tensorflow.keras.layers as layers
from akida import (Model, Convolutional, FullyConnected, SeparableConvolutional,
                   InputData, InputConvolutional, ConvolutionMode, PoolingType,
                   LayerType)
from . import quantization_layers as qlayers


def _get_convolution_mode(str_conv_mode):
    if str_conv_mode == 'same':
        return ConvolutionMode.Same
    return ConvolutionMode.Valid


def _parse_input_data(layer, params):
    params["input_height"] = int(layer.input_shape[1])
    params["input_width"] = int(layer.input_shape[2])
    params["input_channels"] = int(layer.input_shape[3])
    params['name'] = layer.name + "_input"


def _parse_input_conv(layer, params, input_shift):
    if not isinstance(layer, qlayers.QuantizedConv2D):
        raise TypeError(f"First layer {layer.name} must be QuantizedConv2D "
                        "when input_is_image=True. Received layer of type "
                        f"{layer.__class__.__name__}")
    params["input_height"] = int(layer.input_shape[1])
    params["input_width"] = int(layer.input_shape[2])
    params["convolution_mode"] = _get_convolution_mode(layer.padding)
    params["kernel_height"] = layer.kernel_size[0]
    params["kernel_width"] = layer.kernel_size[1]
    params["num_neurons"] = int(layer.kernel.shape[3])
    params["weights_bits"] = layer.quantizer.bitwidth
    params["input_channels"] = int(layer.input_shape[3])
    params["stride_x"] = layer.strides[1]
    params["stride_y"] = layer.strides[0]
    params["name"] = layer.name
    params["padding_value"] = int(input_shift)


def _parse_conv(layer, params):
    if not isinstance(layer, qlayers.QuantizedConv2D):
        raise TypeError(f"Layer {layer.name} was expected to be "
                        "QuantizedConv2D")
    params["convolution_mode"] = _get_convolution_mode(layer.padding)
    params["kernel_height"] = layer.kernel_size[0]
    params["kernel_width"] = layer.kernel_size[1]
    params["num_neurons"] = int(layer.kernel.shape[3])
    params["weights_bits"] = layer.quantizer.bitwidth
    params["stride_x"] = layer.strides[1]
    params["stride_y"] = layer.strides[0]
    params["name"] = layer.name


def _parse_separable_conv(layer, params):
    if not isinstance(layer, qlayers.QuantizedSeparableConv2D):
        raise TypeError(f"Layer {layer.name} was expected to be "
                        "QuantizedSeparableConv2D")
    if layer.quantizer_dw.bitwidth != layer.quantizer.bitwidth:
        raise ValueError(f"Quantized layer {layer.name} must have the same "
                         f"bitwidth for depthwise and pointwise quantizers.")
    params["convolution_mode"] = _get_convolution_mode(layer.padding)
    params["kernel_height"] = layer.kernel_size[0]
    params["kernel_width"] = layer.kernel_size[1]
    # num neurons is set to the number of filters of the depthwise
    params["num_neurons"] = int(layer.pointwise_kernel.shape[3])
    params["weights_bits"] = layer.quantizer.bitwidth
    params["stride_x"] = layer.strides[1]
    params["stride_y"] = layer.strides[0]
    params["name"] = layer.name


def _parse_dense(layer, params):
    if not isinstance(layer, qlayers.QuantizedDense):
        raise TypeError(f"Layer {layer.name} was expected to be "
                        "QuantizedDense")
    params["num_neurons"] = layer.units
    params["weights_bits"] = layer.quantizer.bitwidth
    params["name"] = layer.name


def _parse_max_pooling(layer, params):
    if not isinstance(layer, layers.MaxPooling2D):
        raise TypeError(f"Layer {layer.name} was expected to be MaxPooling2D")
    params["pooling_type"] = PoolingType.Max
    params["pooling_height"] = layer.pool_size[0]
    params["pooling_width"] = layer.pool_size[1]
    params["pooling_stride_y"] = layer.strides[0]
    params["pooling_stride_x"] = layer.strides[1]


def _parse_global_average_pooling(layer, params):
    if not isinstance(layer, layers.GlobalAveragePooling2D):
        raise TypeError(f"Layer {layer.name} was expected to be "
                        "GlobalAveragePooling2D")
    params["pooling_type"] = PoolingType.Average


def _create_akida_layer(layer_type, params):
    """Returns an Akida layer based on the input dictionary containing the
    parameters.

    """
    if layer_type == LayerType.InputData:
        layer_ak = InputData(**params)
    elif layer_type == LayerType.InputConvolutional:
        layer_ak = InputConvolutional(**params)
    elif layer_type == LayerType.Convolutional:
        layer_ak = Convolutional(**params)
    elif layer_type == LayerType.SeparableConvolutional:
        layer_ak = SeparableConvolutional(**params)
    elif layer_type == LayerType.FullyConnected:
        layer_ak = FullyConnected(**params)

    return layer_ak


def generate_model(model_map, input_scaling):
    """Generates an Akida model.

    This function creates an Akida model from the model map: for each Akida
    layer, the Keras parameters are retrieved from the neural, pooling and
    activation layers, and these parameters are then used to instantiate the
    Akida layer.
    Note that this function generates the model but doesn't set weights, fire
    thresholds and fire threshold steps.

    Notes:
        The relationship between Keras and Akida inputs is:
        input_akida = input_factor * input_keras + input_shift
        with input_scaling = (input_factor, input_shift)

    Args:
        model_map (ModelMapping): a model map with the Keras model and the list
            of layer maps.
        input_scaling (2-element tuple): the input factor and shift.

    Returns:
        :obj:`akida.Model`: the generated Akida model.

    """
    layers_keras = model_map.model_keras.layers
    _, input_shift = input_scaling

    model_ak = Model()
    for layer_map in model_map.layer_maps:
        # Get layer params
        layer_type = layer_map.layer_type
        layer = layers_keras[layer_map.index_neural]
        params = {}
        if layer_type == LayerType.InputData:
            _parse_input_data(layer, params)
        elif layer_type == LayerType.InputConvolutional:
            _parse_input_conv(layer, params, input_shift)
        elif layer_type == LayerType.Convolutional:
            _parse_conv(layer, params)
        elif layer_type == LayerType.SeparableConvolutional:
            _parse_separable_conv(layer, params)
        elif layer_type == LayerType.FullyConnected:
            _parse_dense(layer, params)

        # Add pooling params
        if layer_map.index_pool:
            pooling_layer = layers_keras[layer_map.index_pool]
            if isinstance(pooling_layer, layers.MaxPooling2D):
                _parse_max_pooling(pooling_layer, params)
            elif isinstance(pooling_layer, layers.GlobalAveragePooling2D):
                _parse_global_average_pooling(pooling_layer, params)

        # Add activation params
        if layer_map.index_activation:
            activation = layers_keras[layer_map.index_activation]
            if activation.bitwidth not in range(1, 5):
                raise ValueError("Activation bitwidth must be an integer "
                                 "between 1 and 4. Receives bitwidth "
                                 f"{activation.bitwidth} in layer "
                                 f"{activation.name}.")
            params["threshold_fire_bits"] = activation.bitwidth
        elif layer_type != LayerType.InputData:
            params["activations_enabled"] = False

        # Create layer and add it to the Akida model
        layer_ak = _create_akida_layer(layer_type, params)
        model_ak.add(layer_ak)

    return model_ak



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""CNN2SNN objects required to serialize Keras quantized models.
"""

from .quantization_ops import (StdWeightQuantizer, TrainableStdWeightQuantizer,
                               MaxQuantizer, MaxPerAxisQuantizer,
                               ScaleFactorQuantizer)
from .quantization_layers import (QuantizedConv2D, QuantizedSeparableConv2D,
                                  QuantizedDense, ActivationDiscreteRelu,
                                  QuantizedReLU)

cnn2snn_objects = {
    'StdWeightQuantizer': StdWeightQuantizer,
    'TrainableStdWeightQuantizer': TrainableStdWeightQuantizer,
    'MaxQuantizer': MaxQuantizer,
    'MaxPerAxisQuantizer': MaxPerAxisQuantizer,
    'ScaleFactorQuantizer': ScaleFactorQuantizer,
    'QuantizedConv2D': QuantizedConv2D,
    'QuantizedSeparableConv2D': QuantizedSeparableConv2D,
    'QuantizedDense': QuantizedDense,
    'ActivationDiscreteRelu': ActivationDiscreteRelu,
    'QuantizedReLU': QuantizedReLU
}



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Parsing function to generate layers mapping between Keras and Akida.
Two data classes store the mapping: LayerMapping and ModelMapping.

"""
import tensorflow.keras.layers as layers
from akida import LayerType
from . import quantization_layers as qlayers


class LayerMapping:
    """ Creates a layer map of a single Akida layer from Keras layers.

    This data class stores the indices of Keras layers that represent a
    single Akida layer. For example, a 'Convolutional' Akida layer corresponds
    to multiple Keras layers:
    - a Conv2D/QuantizedConv2D layer
    - an optional batch normalization layer
    - an optional pooling layer
    - a ReLU or discrete ReLU activation (optional if last layer)

    Args:
        layer_type (:obj:`akida.LayerType`): the type of the Akida layer.
        index_neural (int): the index of the corresponding Keras
            neural layer.

    """

    def __init__(self, layer_type, index_neural):
        self.layer_type = layer_type
        self.index_neural = index_neural
        self.index_pool = None
        self.index_batchnorm = None
        self.index_activation = None


class ModelMapping:
    """This data class maps a Keras model to a future Akida model (not built yet).

    When an instance of ModelMapping is created, it will generate a list of
    LayerMapping objects mapping the Keras model with a succession of Akida
    layers.
    A check is then performed to ensure that the Keras model is compatible with
    Akida.

    Note:
        Note that no Akida model is generated at this step: only a mapping is
        created.

    """

    def __init__(self, model_keras, layer_maps):
        self.model_keras = model_keras
        self.layer_maps = layer_maps


def _check_layers_data_format(model):
    """Asserts that all layers in the model are 'channels_last'.

    Args:
        model (tf.keras.model): the Keras model to check.
    """

    # Error if 'data_format' is 'channels_first'
    for layer in model.layers:
        if hasattr(layer, 'data_format'):
            if layer.data_format == "channels_first":
                raise RuntimeError("unsupported data format channels_first")


def _check_model_input_output(model):
    """Asserts that model inputs/outputs are supported for conversion.

    The Keras model must have only one input layer and one output layer. The
    input shape must 4-D (N, H, W, C).

    Args:
        model (tf.keras.model): the Keras model to check.
    """

    # Error if multiple inputs
    if len(model.input_names) > 1:
        raise RuntimeError("Model must have only one input layer. Receives "
                           f"inputs {model.input_names}.")

    # Error if multiple outputs
    if len(model.output_names) > 1:
        raise RuntimeError("Model must have only one output layer. Receives "
                           f"outputs {model.output_names}.")

    # Error if input shape is not 4D, i.e. (N, H, W, C)
    input_shape = model.input_shape
    if len(input_shape) != 4:
        err_msg = ("Input shape of model must be 4-D (batch size + 3-D "
                   f"tensors). Receives input shape {input_shape}. ")
        if len(input_shape) == 2:
            err_msg += (
                "If your model begins with a Dense layer, you must "
                "start your model with a Flatten layer and an input shape of "
                f" (1, 1, {input_shape[1]}) instead of {(input_shape[1],)}.")
        raise RuntimeError(err_msg)


def _map_first_layer(model, input_is_image):
    """Map the first layers of the Keras model to an Akida layer.

    To be Akida-compatible, the first layers of the quantized Keras model must
    be:

    - an optional InputLayer.
    - an optional Rescaling layer
    - the first neural layer. If input is image, this layer must be a Conv2D
        with 1 or 3 channels.

    Depending on input_is_image, an InputData or an InputConvolutional is
    returned. The index of the next layer to process is also returned.

    Args:
        model (tf.keras model): the model to parse.
        input_is_image (bool): True if input is an image (8-bit input with 1 or
            3 channels) followed by QuantizedConv2D. Akida model input will be
            InputConvolutional. If False, Akida model input will be InputData.

    Returns:
       :obj:`LayerMapping`: a layer map corresponding to the first Akida layer.
       int: the index of the next layer to process.

    """

    next_layer = 0
    # If first layer is input layer, skip it
    if isinstance(model.layers[0], layers.InputLayer):
        next_layer = 1

    # If next layer is Rescaling layer, skip it
    if isinstance(model.layers[next_layer],
                  layers.experimental.preprocessing.Rescaling):
        next_layer += 1

    # Get first Akida layer
    if not input_is_image:
        layer_ak = LayerMapping(LayerType.InputData, next_layer)
    else:
        # Error if input_is_image=True and first layer is not a Conv2D
        # or input channels is not 1 or 3.
        if (not isinstance(model.layers[next_layer], layers.Conv2D) or
                not model.layers[next_layer].input_shape[-1] in (1, 3)):
            err_msg = (f"With input_is_image=True, first layer "
                       f"'{model.layers[next_layer].name}' must be"
                       f" Conv2D and input shape must have 1 or 3 channels. "
                       f"Receives layer of type "
                       f"{model.layers[next_layer].__class__.__name__} with "
                       f"{model.layers[next_layer].input_shape[-1]} channels.")
            raise RuntimeError(err_msg)
        layer_ak = LayerMapping(LayerType.InputConvolutional, next_layer)
        next_layer += 1

    return layer_ak, next_layer


def _get_akida_layer_type(layer):
    """Returns Akida layer type of a Keras neural layer

    Args:
        layer (:obj:`tf.keras.layers.Layer`): a Keras layer.
    """
    if type(layer) in (layers.Conv2D, qlayers.QuantizedConv2D):
        return LayerType.Convolutional
    if type(layer) in (layers.SeparableConv2D,
                       qlayers.QuantizedSeparableConv2D):
        return LayerType.SeparableConvolutional
    if type(layer) in (layers.Dense, qlayers.QuantizedDense):
        return LayerType.FullyConnected
    return None


def _fill_akida_layer_map(layer_ak, layers_keras, index_layer):
    """Fill Akida layer map with pooling/batchnorm/activation layers.

    This function modifies and fills the `layer_ak` map, that corresponds to
    a block of a neural layer + optional pooling/batchnorm/activation layers.
    For instance, if the `index_layer` of Keras layers (i.e.
    `layers_keras[index_layer]` is a BatchNormalization layer, the Akida layer
    map will keep this index in `layer_ak.index_batchnorm`. This is needed
    later when creating the Akida model and computing the fire thresholds and
    steps.

    Args:
        layer_ak (:obj:`LayerMapping`): the Akida layer map which will be
            filled with pooling/batchnorm/activation Keras layers.
        layers_keras (list): list of Keras layers of the model.
        index_layer (int): index of the Keras layer to read.
    """

    err_msg = ("Two {0} layers were detected related to {1}: {2} and {3}. Only "
               "one {0} layer is supported.")
    layer = layers_keras[index_layer]

    if type(layer) in (layers.MaxPooling2D, layers.GlobalAveragePooling2D):
        if layer_ak.index_pool:
            raise RuntimeError(
                err_msg.format("pooling",
                               layers_keras[layer_ak.index_neural].name,
                               layers_keras[layer_ak.index_pool].name,
                               layer.name))
        layer_ak.index_pool = index_layer
    elif (isinstance(layer, layers.BatchNormalization) or
          layer.__class__.__name__ == "BatchNormalization"):
        if layer_ak.index_batchnorm:
            raise RuntimeError(
                err_msg.format("BatchNormalization",
                               layers_keras[layer_ak.index_neural].name,
                               layers_keras[layer_ak.index_batchnorm].name,
                               layer.name))
        layer_ak.index_batchnorm = index_layer
    elif isinstance(layer, (qlayers.QuantizedActivation, layers.ReLU)):
        if layer_ak.index_activation:
            raise RuntimeError(
                err_msg.format("activation",
                               layers_keras[layer_ak.index_neural].name,
                               layers_keras[layer_ak.index_activation].name,
                               layer.name))
        layer_ak.index_activation = index_layer


def _check_flatten_layer(layers_keras, flatten_index):
    """Asserts that the Flatten layer is supported for conversion.

    A Flatten layer is supported if it is followed by a Dense layer or if it
    is the last layer of the model.

    Args:
        layers_keras (list): the model's Keras layers.
        flatten_index (int): the index of the Flatten layer in the list of
            layers.
    """

    try:
        if type(layers_keras[flatten_index + 1]) in (layers.Dense,
                                                     qlayers.QuantizedDense):
            return
    except IndexError:
        pass
    raise RuntimeError("Flatten layer only supported before a Dense one")


def _check_unsupported_activations(index_unsupported_activations,
                                   last_neural_index, layers_keras):
    """Asserts that unsupported activations are after the last Akida layer.

    Akida does not support other activations than ReLU/QuantizedActivation. The
    only exception is if an unsupported activation is placed after the last
    neural layer. For example, a Sigmoid/Softmax at the end of the model. In
    this case, the conversion is performed but the unsupported activation is
    discarded.

    Args:
        index_unsupported_activations (list): the indices of the unsupported
            activation layers.
        last_neural_index (int): the index of the last neural layer in the
            model.
        layers_keras (list): the model's Keras layers.
    """

    # Check if the unsupported activation is after the last neural layer (i.e.
    # in the last Akida layer
    for i in index_unsupported_activations:
        if i < last_neural_index:
            raise RuntimeError(
                "Activation layers other than ReLU and quantized "
                "activations are not supported before the last "
                "neural layer. Receives activation layer "
                f"'{layers_keras[i].name}' before the last neural"
                f" layer '{layers_keras[last_neural_index].name}'")

        print(f"Warning: the activation layer '{layers_keras[i].name}' "
              "will be discarded at conversion. The outputs of the Akida "
              "model will be the potentials before this activation layer.")


def generate_model_mapping(model, input_is_image):
    """Generates a model map between Keras and Akida models.

    This function returns a model map from a Keras model. The model map
    corresponds to the Akida layers mapped from the Keras layers.

    Args:
        model (tf.keras model): the model to parse.
        input_is_image (bool): True if input is an image (8-bit input with 1 or
            3 channels) followed by QuantizedConv2D. Akida model input will be
            InputConvolutional. If False, Akida model input will be InputData.

    Returns:
       :obj:`ModelMapping`: a model map corresponding to the input Keras model.

    """

    _check_layers_data_format(model)
    _check_model_input_output(model)
    layer_ak, next_layer = _map_first_layer(model, input_is_image)

    # Loop on layers
    neural_layers = (layers.Conv2D, layers.SeparableConv2D, layers.Dense,
                     qlayers.QuantizedConv2D, qlayers.QuantizedSeparableConv2D,
                     qlayers.QuantizedDense)
    ignore_list = (layers.Dropout)
    activation_list = (layers.Activation, layers.Softmax)
    layer_maps = []
    index_unsupported_activations = []

    for i in range(next_layer, len(model.layers)):
        layer = model.layers[i]
        # If this layer is a neural layer, append the current Akida layer and
        # start a new one
        if type(layer) in neural_layers:
            layer_maps.append(layer_ak)
            layer_ak = LayerMapping(_get_akida_layer_type(layer), i)
            continue

        # Pooling + batchnorm + activation layers -> update akida layer map
        if (type(layer) in (layers.MaxPooling2D, layers.GlobalAveragePooling2D,
                            layers.BatchNormalization, layers.ReLU) or
                isinstance(layer, qlayers.QuantizedActivation) or
                layer.__class__.__name__ == 'BatchNormalization'):
            _fill_akida_layer_map(layer_ak, model.layers, i)
        # Allow flatten before a dense layer
        elif isinstance(layer, layers.Flatten):
            _check_flatten_layer(model.layers, i)
        # Check Reshape compatibility
        elif isinstance(layer, layers.Reshape):
            _check_reshape_layer(layer)
        # Get unsupported activation index to check compatibility later
        elif isinstance(layer, activation_list):
            index_unsupported_activations.append(i)
        # Allow some other layers useful in keras but that will be discarded
        # or ignored during conversion
        elif isinstance(layer, ignore_list):
            continue
        else:
            # If you got here it means the layer is not recognised: raise an error.
            raise RuntimeError(f"Layer {layer.name}: unsupported type "
                               f"{layer.__class__.__name__}.")

    # Append last parsed layer if any
    layer_maps.append(layer_ak)

    _check_unsupported_activations(index_unsupported_activations,
                                   layer_maps[-1].index_neural, model.layers)

    return ModelMapping(model, layer_maps)


def _check_batchnorm_compatibility(layer_map, layers_keras):
    """Asserts BatchNormalization layer is compatible for conversion.

    Four checks are performed here:
    - BN must be applied on the last axis of the input tensor.
    - gammas must not be zero.
    - gammas must be positive if a MaxPool2D layer is placed before BN.
    - BN must be placed before the activation.

    Args:
       layer_map (:obj:`LayerMapping`): the layer map to check.
       layers_keras (list): the model's Keras layers.
    """

    layer_neural = layers_keras[layer_map.index_neural]

    if layer_map.index_batchnorm:
        # Raise error if BatchNormalization 'axis' is different from the last
        # dimension. The 'axis' parameter is a list containing the axes on what
        # the batch normalization is applied.
        layer_BN = layers_keras[layer_map.index_batchnorm]
        if (len(layer_BN.axis) != 1 or
                layer_BN.axis[0] != len(layer_BN.input_shape) - 1):
            raise RuntimeError(f"The BatchNormalization layer "
                               f"{layer_BN.name} must be applied on the "
                               f"last axis. Receives {layer_BN.axis}.")
        # Raise error if a gamma is zero.
        gammas = layer_BN.get_weights()[0]
        if (gammas == 0).any():
            raise RuntimeError(
                f"The BatchNormalization layer {layer_BN.name} has at least"
                f" one gamma equal to zero. This case is not supported.")
        # Raise error if BatchNormalization has at least one negative gamma
        # and if a MaxPool2D layer is placed before it.
        if (gammas <= 0).any() and layer_map.index_pool:
            layer_pool = layers_keras[layer_map.index_pool]
            if (isinstance(layer_pool, layers.MaxPool2D) and
                    layer_map.index_pool < layer_map.index_batchnorm):
                raise RuntimeError(
                    f"The BatchNormalization layer {layer_BN.name} has at "
                    f"least one negative gamma and a MaxPool2D layer is "
                    f"placed before it. This case is not supported.")

    # Raise error if BatchNormalization is placed after the activation
    if (layer_map.index_batchnorm and layer_map.index_activation and
            layer_map.index_batchnorm > layer_map.index_activation):
        raise RuntimeError(f"In the layer {layer_neural.name}, the batch "
                           "normalization layer must be placed before "
                           "the activation.")


def _check_pooling_compatibility(layer_map, layers_keras):
    """Asserts pooling layer is compatible for conversion.

    Two checks are performed here:
    - a global average pooling must be placed before the activation layer
    - the padding of MaxPool2D must be the same as the padding of neural layer

    Args:
       layer_map (:obj:`LayerMapping`): the layer map to check.
       layers_keras (list): the model's Keras layers.
    """

    layer_neural = layers_keras[layer_map.index_neural]

    # Raise error if GlobalAvgPool2D is placed after the activation
    if (layer_map.index_pool and layer_map.index_activation and isinstance(
            layers_keras[layer_map.index_pool], layers.GlobalAvgPool2D) and
            layer_map.index_pool > layer_map.index_activation):
        raise RuntimeError(f"In the layer {layer_neural.name}, the global "
                           "average pooling layer must be placed before "
                           "the activation.")

    # Raises error if the padding of MaxPool2D is different from the padding
    # of the neural processing layer.
    if layer_map.index_pool:
        layer_pool = layers_keras[layer_map.index_pool]
        if (isinstance(layer_pool, layers.MaxPool2D) and
                layer_neural.padding != layer_pool.padding):
            raise RuntimeError(f"Pooling layer {layer_pool.name} (padding: "
                               f"{layer_pool.padding}) must have the same "
                               f"padding as {layer_neural.name} (padding: "
                               f"{layer_neural.padding}).")


def _check_dense_compatibility(layer_map, layers_keras):
    """Asserts Dense layer is compatible for conversion.

    One check is performed here:
    - input shape must be (bs, N) or (bs, 1, 1, N) (bs is the batch size).

    Args:
       layer_map (:obj:`LayerMapping`): the layer map to check.
       layers_keras (list): the model's Keras layers.
    """

    layer_neural = layers_keras[layer_map.index_neural]

    # Raises error if Dense input shape is incorrect: supported
    # shapes are (N,) and (1, 1, N). Remember input_shape has the batch
    # size as first element of tuple.
    if isinstance(layer_neural, layers.Dense):
        valid = (  # Input shape is (N,)
            len(layer_neural.input_shape) == 2 or
            # Input shape is (1, 1, N)
            (len(layer_neural.input_shape) == 4 and
             layer_neural.input_shape[1] == 1 and
             layer_neural.input_shape[2] == 1))
        if not valid:
            raise RuntimeError("The Dense layer "
                               f"{layer_neural.name} must have an input "
                               "shape of (N,). Receives "
                               f"{layer_neural.input_shape[1:]}.")


def check_mapping_compatibility(model_map):
    """Checks whether the future model will be compatible with Akida.

    This function must mainly test the incompatibities due to the Keras layers
    and the order of the layers (parameters of the quantized layers have
    already been tested at their creation).

    Args:
        :obj:`ModelMapping`: a model map corresponding to the Keras model to
            check
    """
    layers_keras = model_map.model_keras.layers

    # Error if hidden layer without activation
    for layer_map in model_map.layer_maps[:-1]:
        if layer_map.layer_type != LayerType.InputData \
                and not layer_map.index_activation:
            raise RuntimeError("No activation layer detected with layer "
                               f"{layers_keras[layer_map.index_neural].name}. "
                               "Activation is required in hidden layers.")

    for layer_map in model_map.layer_maps:
        _check_batchnorm_compatibility(layer_map, layers_keras)
        _check_pooling_compatibility(layer_map, layers_keras)
        _check_dense_compatibility(layer_map, layers_keras)


def _check_reshape_layer(layer):
    """This function checks if the reshape layer is supported.

    In the cnn2snn conversion, a Reshape layer can only be used to transform
    a tensor of shape (N,) to a tensor of shape (1, 1, N), and vice-versa.

    Note that the 'input_shape' and 'output_shape' parameters of a layer has
    the batch size as first element:
        input_shape = (batch_size,) + input_tensor_shape
    The batch size is ignored in the following function.
    """
    in_shape = layer.input_shape
    out_shape = layer.output_shape

    valid = ((  # Reshape from (1,1,N) to (N,)
        len(in_shape) == 4 and in_shape[1] == 1 and in_shape[2] == 1 and
        len(out_shape) == 2 and out_shape[1] == in_shape[3]) or
             # Reshape from (N,) to (1,1,N)
             (len(in_shape) == 2 and len(out_shape) == 4 and
              out_shape[1] == 1 and out_shape[2] == 1 and
              out_shape[3] == in_shape[1]) or
             # Useless Reshape, from X to X
             (in_shape == out_shape))

    if not valid:
        raise RuntimeError(f"The Reshape layer {layer.name} can only be used "
                           "to transform a tensor of shape (N,) to a tensor of "
                           "shape (1, 1, N), and vice-versa. Receives "
                           f"input_shape {in_shape[1:]} and output_shape "
                           f"{out_shape[1:]}.")


def check_model_compatibility(model_keras, input_is_image=True):
    r"""Checks if a Keras model is compatible for cnn2snn conversion.

    This function doesn't convert the Keras model to an Akida model
    but only checks if the model design is compatible. The checks are performed
    at two different levels:

        1. Some checks are done when the Keras model is scanned, during the
           generation of the model map.
        2. Other checks are then done based on the model map.

    Note that this function doesn't check if the quantization bitwidths (weights
    or activations) are supported by the Akida Execution Engine or by the Akida
    NSoC.

    **1. How to build a compatible Keras quantized model?**

    The following lines give details and constraints on how to build a Keras
    model compatible for the conversion to an Akida model.


    **2. General information about layers**

    An Akida layer must be seen as a block of Keras layers starting with a
    processing layer (Conv2D, SeparableConv2D,
    Dense). All blocks of Keras layers except the last block must have
    exactly one activation layer (ReLU or ActivationDiscreteRelu). Other
    optional layers can be present in a block such as a pooling layer or a
    batch normalization layer.
    Here are all the supported Keras layers for an Akida-compatible model:

    - Processing layers:

      - tf.keras Conv2D/SeparableConv2D/Dense
      - cnn2snn QuantizedConv2D/QuantizedSeparableConv2D/QuantizedDense

    - Activation layers:

      - tf.keras ReLU
      - cnn2snn ActivationDiscreteRelu
      - any increasing activation function (only for the last block of layers)
        such as softmax, sigmoid set as last layer. This layer must derive from
        tf.keras.layers.Activation, and it will be removed during conversion to
        an Akida model.

    - Pooling layers:

      - MaxPool2D
      - GlobalAvgPool2D

    - BatchNormalization
    - Dropout
    - Flatten
    - Input
    - Reshape

    Example of a block of Keras layers::

              ----------
              | Conv2D |
              ----------
                  ||
                  \/
        ----------------------
        | BatchNormalization |
        ----------------------
                  ||
                  \/
             -------------
             | MaxPool2D |
             -------------
                  ||
                  \/
       --------------------------
       | ActivationDiscreteRelu |
       --------------------------


    **3. Constraints about inputs**

    An Akida model can accept two types of inputs: sparse events or 8-bit
    images. Whatever the input type, the Keras inputs must respect the
    following relation:

        input_akida = scale * input_keras + shift

    where the Akida inputs must be positive integers, the input scale must be
    a float value and the input shift must be an integer. In other words,
    scale * input_keras must be integers.

    Depending on the input type:

    - if the inputs are events (sparse), the first layer of the Keras model can
      be any processing layer. The input shift must be zero.
    - if the inputs are images, the first layer must be a Conv2D
      layer.


    **4. Constraints about layers' parameters**

    To be Akida-compatible, the Keras layers must observe the following rules:

    - all layers with the 'data_format' parameter must be 'channels_last'
    - all processing quantized layers and ActivationDiscreteRelu must have a
      valid quantization bitwidth
    - a Dense layer must have an input shape of (N,) or (1, 1, N)
    - a BatchNormalization layer must have 'axis' set to -1 (default)
    - a BatchNormalization layer cannot have negative gammas
    - Reshape layers can only be used to transform a tensor of shape (N,) to a
      tensor of shape (1, 1, N), and vice-versa
    - only one pooling layer can be used in each block
    - a MaxPool2D layer must have the same 'padding' as the corresponding
      processing quantized layer

    **5. Constraints about the order of layers**

    To be Akida-compatible, the order of Keras layers must observe the following
    rules:

    - a block of Keras layers must start with a processing quantized layer
    - where present, a BatchNormalization/GlobalAvgPool2D layer must be placed
      before the activation
    - a Flatten layer can only be used before a Dense layer
    - an Activation layer other than ActivationDiscreteRelu can only be used
      in the last layer


    Args:
        model (tf.keras model): the model to parse.
        input_is_image (bool, optional): True if input is an image (8-bit input
            with 1 or 3 channels) followed by QuantizedConv2D. Akida model
            input will be InputConvolutional. If False, Akida model input will
            be InputData. (Default value = True)
    """
    try:
        model_map = generate_model_mapping(model_keras, input_is_image)
        check_mapping_compatibility(model_map)
        return True
    except RuntimeError as e:
        print(
            "The Keras quantized model is not compatible for a conversion "
            "to an Akida model:\n", str(e))
        return False



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Quantized layers API: neural quantized layers and quantized activations"""

# Tensorflow imports
import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras import layers
from tensorflow.python.keras.layers.ops import core as core_ops  # pylint: disable=no-name-in-module
from tensorflow.keras.utils import serialize_keras_object
from tensorflow.python.keras.utils import conv_utils  # pylint: disable=no-name-in-module
from tensorflow.python.ops import nn  # pylint: disable=no-name-in-module
from tensorflow.python.ops import nn_ops  # pylint: disable=no-name-in-module
from .quantization_ops import get as get_quantizer
from .quantization_ops import ceil_through


def _check_unsupported_args(kwargs, unsupported_args):
    """Raises error if unsupported argument are present in kwargs.

    For now, 4 arguments are unsupported: 'data_format', 'activation',
        'dilation_rate', 'depth_mutiplier'.

    Args:
        kwargs (dictionary): keyword arguments to check.
        unsupported_args: list of unsupported arguments.

    """
    for kwarg in kwargs:
        if kwarg in unsupported_args:
            raise TypeError("Unsupported argument in quantized layers:", kwarg)


class QuantizedConv2D(layers.Conv2D):
    """A quantization-aware Keras convolutional layer.

    Inherits from Keras Conv2D layer, applying a quantization on weights during
    the forward pass.

    Args:
        filters (int): the number of filters.
        kernel_size (tuple of integer): the kernel spatial dimensions.
        quantizer (:obj:`cnn2snn.WeightQuantizer`): the quantizer
            to apply during the forward pass.
        strides (integer, or tuple of integers, optional): strides of the
            convolution along height and width.
        padding (str, optional): one of 'valid' or 'same'.
        use_bias (boolean, optional): whether the layer uses a bias vector.
        kernel_initializer (str, or a :obj:`tf.keras.initializer`, optional):
            initializer for the weights matrix.
        bias_initializer (str, or a :obj:`tf.keras.initializer`, optional):
            initializer for the bias vector.
        kernel_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
            regularization applied to the weights.
        bias_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
            regularization applied to the bias.
        activity_regularizer (str, or a :obj:`tf.keras.regularizer`,
            optional): regularization applied to the output of the layer.
        kernel_constraint (str, or a :obj:`tf.keras.constraint`, optional):
            constraint applied to the weights.
        bias_constraint (str, or a :obj:`tf.keras.constraint`, optional):
            constraint applied to the bias.

    """
    unsupported_args = {
        'data_format': 'channels_last',
        'activation': 'linear',
        'dilation_rate': (1, 1),
        'groups': 1
    }

    def __init__(self,
                 filters,
                 kernel_size,
                 quantizer,
                 strides=(1, 1),
                 padding='valid',
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):

        _check_unsupported_args(kwargs, self.unsupported_args)
        self.quantizer = get_quantizer(quantizer)
        super().__init__(filters=filters,
                         kernel_size=kernel_size,
                         strides=strides,
                         padding=padding,
                         use_bias=use_bias,
                         kernel_initializer=kernel_initializer,
                         bias_initializer=bias_initializer,
                         kernel_regularizer=kernel_regularizer,
                         bias_regularizer=bias_regularizer,
                         activity_regularizer=activity_regularizer,
                         kernel_constraint=kernel_constraint,
                         bias_constraint=bias_constraint,
                         **kwargs)

    def call(self, inputs):
        """Evaluates input Tensor.

        This applies the quantization on weights, then evaluates the input
        Tensor and produces the output Tensor.

        Args:
            inputs(:obj:`tensorflow.Tensor`): input Tensor.

        Returns:
            :obj:`tensorflow.Tensor`: output Tensor.

        """
        outputs = self._convolution_op(inputs,
                                       self.quantizer.quantize(self.kernel))

        if self.use_bias:
            # Handle multiple batch dimensions.
            output_rank = outputs.shape.rank
            if output_rank is not None and output_rank > 2 + self.rank:

                def _apply_fn(o):
                    return nn.bias_add(o,
                                       self.bias,
                                       data_format=self._tf_data_format)

                outputs = nn_ops.squeeze_batch_dims(outputs,
                                                    _apply_fn,
                                                    inner_rank=self.rank + 1)

            else:
                outputs = nn.bias_add(outputs,
                                      self.bias,
                                      data_format=self._tf_data_format)

        return outputs

    def get_config(self):
        config = super().get_config()
        config['quantizer'] = serialize_keras_object(self.quantizer)
        for kwarg in self.unsupported_args:
            config.pop(kwarg, None)
        return config


class QuantizedDepthwiseConv2D(layers.DepthwiseConv2D):
    """A quantization-aware Keras depthwise convolutional layer.

    Inherits from Keras DepthwiseConv2D layer, applying a quantization on
    weights during the forward pass.

    Args:
        kernel_size (a tuple of integer): the kernel spatial dimensions.
        strides (integer, or tuple of integers, optional): strides of the
            convolution along height and width.
        quantizer (:obj:`cnn2snn.WeightQuantizer`): the quantizer
            to apply during the forward pass.
        padding (str, optional): One of 'valid' or 'same'.
        use_bias (boolean, optional): whether the layer uses a bias vector.
        depthwise_initializer (str, or a :obj:`tf.keras.initializer`, optional):
            initializer for the weights matrix.
        bias_initializer (str, or a :obj:`tf.keras.initializer`, optional):
            initializer for the bias vector.
        depthwise_regularizer (str, or a :obj:`tf.keras.initializer`, optional):
            regularization applied to the weights.
        bias_regularizer (str, or a :obj:`tf.keras.initializer`, optional):
            regularization applied to the bias.
        activity_regularizer (str, or a :obj:`tf.keras.regularizer`,
            optional): regularization applied to the output of the layer.
        depthwise_constraint (str, or a :obj:`tf.keras.initializer`, optional):
            constraint applied to the weights.
        bias_constraint (str, or a :obj:`tf.keras.initializer`, optional):
            constraint applied to the bias.

    """
    unsupported_args = {
        'data_format': 'channels_last',
        'activation': 'linear',
        'depth_multiplier': 1
    }

    def __init__(self,
                 kernel_size,
                 quantizer,
                 strides=(1, 1),
                 padding='valid',
                 use_bias=True,
                 depthwise_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 depthwise_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 depthwise_constraint=None,
                 bias_constraint=None,
                 **kwargs):

        _check_unsupported_args(kwargs, self.unsupported_args)
        self.quantizer = get_quantizer(quantizer)
        super().__init__(kernel_size=kernel_size,
                         strides=strides,
                         padding=padding,
                         use_bias=use_bias,
                         depthwise_initializer=depthwise_initializer,
                         bias_initializer=bias_initializer,
                         depthwise_regularizer=depthwise_regularizer,
                         bias_regularizer=bias_regularizer,
                         activity_regularizer=activity_regularizer,
                         depthwise_constraint=depthwise_constraint,
                         bias_constraint=bias_constraint,
                         **kwargs)

    def call(self, inputs):
        """Evaluates input Tensor.

        This applies the quantization on weights, then evaluates the input
        Tensor and produces the output Tensor.

        Args:
            inputs (:obj:`tensorflow.Tensor`): input Tensor.

        Returns:
            :obj:`tensorflow.Tensor`: output Tensor.

        """
        # We don't support biases
        return K.depthwise_conv2d(inputs,
                                  self.quantizer.quantize(
                                      self.depthwise_kernel),
                                  strides=self.strides,
                                  padding=self.padding,
                                  dilation_rate=self.dilation_rate,
                                  data_format=self.data_format)

    def get_config(self):
        config = super().get_config()
        config['quantizer'] = serialize_keras_object(self.quantizer)
        for kwarg in self.unsupported_args:
            config.pop(kwarg, None)
        return config


class QuantizedDense(layers.Dense):
    """A quantization-aware Keras dense layer.

    Inherits from Keras Dense layer, applying a quantization on weights during
    the forward pass.

    Args:
        units (int): the number of neurons.
        use_bias (boolean, optional): whether the layer uses a bias vector.
        quantizer (:obj:`cnn2snn.WeightQuantizer`): the quantizer
            to apply during the forward pass.
        kernel_initializer (str, or a :obj:`tf.keras.initializer`, optional):
            initializer for the weights matrix.
        bias_initializer (str, or a :obj:`tf.keras.initializer`, optional):
            initializer for the bias vector.
        kernel_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
            regularization applied to the weights.
        bias_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
            regularization applied to the bias.
        activity_regularizer (str, or a :obj:`tf.keras.regularizer`,
            optional): regularization applied to the output of the layer.
        kernel_constraint (str, or a :obj:`tf.keras.constraint`, optional):
            constraint applied to the weights.
        bias_constraint (str, or a :obj:`tf.keras.constraint`, optional):
            constraint applied to the bias.

    """
    unsupported_args = {'activation': 'linear'}

    def __init__(self,
                 units,
                 quantizer,
                 use_bias=True,
                 kernel_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 kernel_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 kernel_constraint=None,
                 bias_constraint=None,
                 **kwargs):

        _check_unsupported_args(kwargs, self.unsupported_args)
        self.quantizer = get_quantizer(quantizer)
        super().__init__(units=units,
                         use_bias=use_bias,
                         kernel_initializer=kernel_initializer,
                         bias_initializer=bias_initializer,
                         kernel_regularizer=kernel_regularizer,
                         bias_regularizer=bias_regularizer,
                         activity_regularizer=activity_regularizer,
                         kernel_constraint=kernel_constraint,
                         bias_constraint=bias_constraint,
                         **kwargs)

    def call(self, inputs):
        """Evaluates input Tensor.

        This applies the quantization on weights, then evaluates the input
        Tensor and produces the output Tensor.

        Args:
            inputs (:obj:`tensorflow.Tensor`): input Tensor.

        Returns:
            :obj:`tensorflow.Tensor`: output Tensor.

        """
        kernel = self.quantizer.quantize(self.kernel)
        return core_ops.dense(inputs,
                              kernel,
                              self.bias,
                              self.activation,
                              dtype=self._compute_dtype_object)

    def get_config(self):
        config = super().get_config()
        config['quantizer'] = serialize_keras_object(self.quantizer)
        for kwarg in self.unsupported_args:
            config.pop(kwarg, None)
        return config


class QuantizedSeparableConv2D(layers.SeparableConv2D):
    """A quantization-aware Keras separable convolutional layer.

    Inherits from Keras SeparableConv2D layer, applying a quantization on
    weights during the forward pass.

    Creates a quantization-aware separable convolutional layer.

    Args:
        filters (int): the number of filters.
        kernel_size (tuple of integer): the kernel spatial dimensions.
        quantizer (:obj:`cnn2snn.WeightQuantizer`): the quantizer to apply
            during the forward pass.
        quantizer_dw (:obj:`cnn2snn.WeightQuantizer`, optional): the
            depthwise quantizer to apply during the forward pass.
        strides (integer, or tuple of integers, optional): strides of the
            convolution along height and width.
        padding (str, optional): One of 'valid' or 'same'.
        use_bias (boolean, optional): Whether the layer uses a bias vector.
        depthwise_initializer (str, or a :obj:`tf.keras.initializer`, optional):
            initializer for the depthwise kernel.
        pointwise_initializer (str, or a :obj:`tf.keras.initializer`, optional):
            initializer for the pointwise kernel.
        bias_initializer (str, or a :obj:`tf.keras.initializer`, optional):
            initializer for the bias vector.
        depthwise_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
            regularization applied to the depthwise kernel.
        pointwise_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
            regularization applied to the pointwise kernel.
        bias_regularizer (str, or a :obj:`tf.keras.regularizer`, optional):
            regularization applied to the bias.
        activity_regularizer (str, or a :obj:`tf.keras.regularizer`,
            optional): regularization applied to the output of the layer.
        depthwise_constraint (str, or a :obj:`tf.keras.constraint`, optional):
            constraint applied to the depthwise kernel.
        pointwise_constraint (str, or a :obj:`tf.keras.constraint`, optional):
            constraint applied to the pointwise kernel.
        bias_constraint (str, or a :obj:`tf.keras.constraint`, optional):
            constraint applied to the bias.

    """
    unsupported_args = {
        'data_format': 'channels_last',
        'activation': 'linear',
        'dilation_rate': (1, 1),
        'depth_multiplier': 1
    }

    def __init__(self,
                 filters,
                 kernel_size,
                 quantizer,
                 quantizer_dw=None,
                 strides=(1, 1),
                 padding='valid',
                 use_bias=True,
                 depthwise_initializer='glorot_uniform',
                 pointwise_initializer='glorot_uniform',
                 bias_initializer='zeros',
                 depthwise_regularizer=None,
                 pointwise_regularizer=None,
                 bias_regularizer=None,
                 activity_regularizer=None,
                 depthwise_constraint=None,
                 pointwise_constraint=None,
                 bias_constraint=None,
                 **kwargs):
        # pylint: disable=too-many-locals

        _check_unsupported_args(kwargs, self.unsupported_args)
        self.quantizer = get_quantizer(quantizer)
        if quantizer_dw is None:
            # If no depthwise quantizer provided, use the pointwise quantizer
            # Note: this is compatible with legacy models
            self.quantizer_dw = self.quantizer.__class__.from_config(
                self.quantizer.get_config())
        else:
            self.quantizer_dw = get_quantizer(quantizer_dw)

        super().__init__(filters=filters,
                         kernel_size=kernel_size,
                         strides=strides,
                         padding=padding,
                         use_bias=use_bias,
                         depthwise_initializer=depthwise_initializer,
                         pointwise_initializer=pointwise_initializer,
                         bias_initializer=bias_initializer,
                         depthwise_regularizer=depthwise_regularizer,
                         pointwise_regularizer=pointwise_regularizer,
                         bias_regularizer=bias_regularizer,
                         activity_regularizer=activity_regularizer,
                         depthwise_constraint=depthwise_constraint,
                         pointwise_constraint=pointwise_constraint,
                         bias_constraint=bias_constraint,
                         **kwargs)

    def call(self, inputs):
        """Evaluates input Tensor.

        This applies the quantization on weights, then evaluates the input
        Tensor and produces the output Tensor.

        Args:
            inputs (:obj:`tensorflow.Tensor`): input Tensor.

        Returns:
            :obj:`tensorflow.Tensor`: a Tensor.

        """
        strides = (1,) + self.strides + (1,)
        outputs = nn.separable_conv2d(
            inputs,
            self.quantizer_dw.quantize(self.depthwise_kernel),
            self.quantizer.quantize(self.pointwise_kernel),
            strides=strides,
            padding=self.padding.upper(),
            rate=self.dilation_rate,
            data_format=conv_utils.convert_data_format(self.data_format,
                                                       ndim=4))

        if self.use_bias:
            outputs = nn.bias_add(outputs,
                                  self.bias,
                                  data_format=conv_utils.convert_data_format(
                                      self.data_format, ndim=4))

        return outputs

    def get_config(self):
        config = super().get_config()
        config['quantizer_dw'] = serialize_keras_object(self.quantizer_dw)
        config['quantizer'] = serialize_keras_object(self.quantizer)
        for kwarg in self.unsupported_args:
            config.pop(kwarg, None)
        return config


class QuantizedActivation(layers.Layer):
    """Base class for quantized activation layers.

    This base class must be overloaded as well as the three @property
    functions:

    - `threshold`
    - `step_height`
    - `step_width`

    These @property functions must return TensorFlow objects (e.g. tf.Tensor
    or tf.Variable) of scalar values. The `.numpy()` method must be callable on
    them. They can be fixed at initialization or can be trainable variables.

    The CNN2SNN toolkit only support linear quantized activation as defined in
    the `quantized_activation` function.

    The bitwidth defines the number of quantization levels on which the
    activation will be quantized. For instance, a 4-bit quantization gives
    15 activation levels. More generally, a n-bit quantization gives 2^n-1
    levels.

    Args:
        bitwidth (int): the quantization bitwidth

    """

    def __init__(self, bitwidth, **kwargs):
        if bitwidth <= 0:
            raise ValueError("Activation 'bitwidth' must be greater than zero."
                             f" Receives 'bitwidth' {bitwidth}.")

        self.bitwidth_ = bitwidth
        self.levels = 2.**bitwidth - 1
        super().__init__(**kwargs)

    def quantized_activation(self, x):
        """Evaluates the quantized activations for the specified input Tensor.

        The quantization is defined by three parameters:

        - the activation threshold, 'threshold'
        - the quantization step size, 'step_height'
        - the step width, 'step_width'

        For any potential x, the activation output is as follows:

        - if x <= threshold, activation is zero
        - if threshold < x <= threshold + step_width, activation is
          step_height
        - if threshold + step_width < x <= threshold + 2*step_width,
          activation is 2*step_height
        - and so on...
        - if x > threshold + levels*step_width, activation is
          levels*step_height

        Args:
            x (:obj:`tensorflow.Tensor`): the input values.

        """
        act = ceil_through((x - self.threshold) / self.step_width)
        clip_act = tf.clip_by_value(act, 0, self.levels)
        return self.step_height * clip_act

    def call(self, inputs, **kwargs):
        return self.quantized_activation(inputs)

    @property
    def bitwidth(self):
        """Returns the bitwidth of the quantized activation"""
        return self.bitwidth_

    @property
    def threshold(self):
        """Returns the activation threshold"""
        raise NotImplementedError()

    @property
    def step_height(self):
        """Returns the step height of the quantized activation"""
        raise NotImplementedError()

    @property
    def step_width(self):
        """Returns the step width of the quantized activation"""
        raise NotImplementedError()

    def get_config(self):
        config = super().get_config()
        config.update({'bitwidth': self.bitwidth_})
        return config


class ActivationDiscreteRelu(QuantizedActivation):
    """A discrete ReLU Keras Activation.

    Activations will be quantized and will have 2^bitwidth values in the range
    [0,6].

    Args:
        bitwidth (int): the activation bitwidth.

    """

    def __init__(self, bitwidth=1, **kwargs):
        super().__init__(bitwidth, **kwargs)

        relumax = min(self.levels, 6)
        self.step_height_ = tf.constant(relumax / self.levels)

    @property
    def threshold(self):
        return 0.5 * self.step_height_

    @property
    def step_height(self):
        return self.step_height_

    @property
    def step_width(self):
        return self.step_height_


class QuantizedReLU(QuantizedActivation):
    """A Trainable Quantized ReLU Keras Activation.

    Activations will be clipped to a trainable range, and quantized to a number
    of values defined by the bitwidth: N = (2^bitwidth - 1) values plus zero

    More specifically, this class uses two trainable variables:

    - threshold represents the lower bound of the activation range,
    - step_height represents the step between two quantized activation values.

    The activation range is therefore [threshold, tN_k], with:

        tN_k = threshold + N * step_height = (2^bitwidth - 1) * step_height

    In other words:

    - inputs below threshold will result in no activation
    - inputs between threshold and threshold + tN_k will be ceiled to the nearest
      threshold + n * step_height, and result in a activation of n * step_height
    - inputs above threshold + tN_k will result in a activation of N * step_height

    Args:
        bitwidth (int): the activation bitwidth.

    """

    def __init__(self, bitwidth=1, **kwargs):
        super().__init__(bitwidth, **kwargs)

        # To be compatible with models trained with ActivationDiscreteRelu, we
        # keep the same rule for the initial value of the activation range
        # relumax and we deduce the initial values of the threshold and
        # step_height from it
        relumax = min(self.levels, 6)
        step_height_initializer = relumax / self.levels
        # Create lower bound variable and initialize it
        threshold_initializer = step_height_initializer / 2.
        self.threshold_ = tf.Variable(name=f"{self.name}/threshold",
                                      initial_value=threshold_initializer,
                                      dtype=tf.float32,
                                      trainable=True)
        # We will actually train the rescaled step, whose value is more
        # consistent between different bitwidth values
        step_initializer = (2.**self.bitwidth) * step_height_initializer / 16
        self.step_ = tf.Variable(name=f"{self.name}/step",
                                 initial_value=step_initializer,
                                 dtype=tf.float32,
                                 trainable=True)

    @property
    def step_height(self):
        return self.step_width

    @property
    def threshold(self):
        return self.threshold_

    @property
    def step_width(self):
        return 16 * self.step_ / 2.**self.bitwidth



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert weights from Keras to Akida.

"""
from collections import namedtuple
import numpy as np
import tensorflow as tf
import tensorflow.keras.layers as layers
from akida import LayerType

KerasLayerParams = namedtuple('KerasLayerParams', [
    'bias', 'act_threshold', 'act_step', 'gamma_bn', 'sigma_bn', 'beta_bn',
    'mu_bn', 'global_avg_pool_factor'
])


def set_weights_thresholds(model_map, model_ak, input_scaling):
    """Converts and sets weights, fire thresholds and threshold steps to
    an Akida model from a model map.

    Args:
        model_map (ModelMapping): a model map that represents the Akida layers
            mapped to a Keras model
        model_ak (:obj:`akida.Model`): an Akida model that has been generated from the
            corresponding model map
        input_scaling (2-element tuple): the input factor and shift

    """
    layers_keras = model_map.model_keras.layers

    # Loop over the Akida layers
    current_input_scaling = input_scaling
    for i in range(len(model_map.layer_maps)):
        layer_map = model_map.layer_maps[i]
        layer_ak = model_ak.get_layer(i)

        if layer_map.layer_type == LayerType.InputData:
            continue

        # Convert and set weights
        layer_k = layers_keras[layer_map.index_neural]
        weights, weights_scale_factor = \
            convert_layer_weights(layer_map.layer_type, layer_k,
                                  layer_ak.input_dims)

        # Convert and set fire thresholds and steps
        th_fire, th_step = convert_layer_thresholds(layer_map, layers_keras,
                                                    current_input_scaling,
                                                    weights_scale_factor,
                                                    weights)

        # Handle negative BN gammas by multiplying weights and thresholds by -1
        neg_gammas = (th_step < 0)
        if np.any(neg_gammas):
            if layer_map.layer_type == LayerType.SeparableConvolutional:
                weights[1][..., neg_gammas] = -weights[1][..., neg_gammas]
            else:
                weights[..., neg_gammas] = -weights[..., neg_gammas]
            th_fire[neg_gammas] = -th_fire[neg_gammas]
            th_step[neg_gammas] = -th_step[neg_gammas]

        if layer_map.layer_type == LayerType.SeparableConvolutional:
            layer_ak.set_variable('weights', weights[0])
            layer_ak.set_variable('weights_pw', weights[1])
        else:
            layer_ak.set_variable('weights', weights)

        layer_ak.set_variable('threshold_fire', th_fire)
        layer_ak.set_variable("threshold_fire_step", th_step)
        if layer_map.index_activation:
            current_input_scaling = (1 / \
                layers_keras[layer_map.index_activation].step_height.numpy(), 0)


def convert_layer_weights(layer_type, layer_k, input_dims=None):
    """Converts weights for an Akida layer.

    This function returns the converted weights for any Akida layer,
    according to the Akida layer type and the Keras neural layer.

    Args:
        layer_type (str): type of the layer Akida.
        layer_k (:obj:`tf.keras.Layer`): the Keras layer from which the weights
            will be converted.
        input_dims (tuple): input dimensions of the Keras layer. This argument
            is only required for a fullyConnected Akida layer.

    Returns:
        tuple: the converted weights and the weights scale factor from
            quantization.

    """
    if layer_type == LayerType.InputConvolutional:
        return _convert_conv_weights(layer_k, input_conv=True)
    if layer_type == LayerType.Convolutional:
        return _convert_conv_weights(layer_k)
    if layer_type == LayerType.SeparableConvolutional:
        return _convert_separable_conv_weights(layer_k)
    if layer_type == LayerType.FullyConnected:
        return _convert_dense_weights(layer_k, input_dims)
    assert False, "The layer type is unmanaged."
    return tuple()


def _convert_conv_weights(layer, input_conv=False):
    # Quantize weights
    weights = layer.get_weights()[0]
    wq, delta = quantize_weights(layer.quantizer, weights)

    # Transpose weights to get from Keras HWCN to Akida WHCN
    # and multiply by the weights scale factor.
    wq_akida = wq.transpose((1, 0, 2, 3)) * delta
    wq_akida = np.round(wq_akida).astype(np.int8)

    # Flip W and H dimensions for conv. kernels (not input conv.)
    if not input_conv:
        wq_akida = np.flip(wq_akida, axis=[0, 1])

    return wq_akida, delta


def _convert_separable_conv_weights(layer):
    # Quantize depthwise weights
    weights = layer.get_weights()[0]
    wq_dw, delta_dw = quantize_weights(layer.quantizer_dw, weights)
    w_ak_dw = wq_dw.transpose((1, 0, 2, 3)) * delta_dw
    w_ak_dw = np.round(w_ak_dw).astype(np.int8)
    w_ak_dw = np.flip(w_ak_dw, axis=[0, 1])

    # Quantize pointwise weights
    weights_pw = layer.get_weights()[1]
    wq_pw, delta_pw = quantize_weights(layer.quantizer, weights_pw)
    # Pointwise weights in Keras have HWCN format and H = W = 1. This
    # makes the conversion to Akida's NCHW trivial.
    w_ak_pw = wq_pw * delta_pw
    w_ak_pw = np.round(w_ak_pw).astype(np.int8)

    # For separable, the weights scale factor can be seen as the product of the
    # DW and PW scale factors.
    delta = delta_dw * delta_pw
    return (w_ak_dw, w_ak_pw), delta


def _convert_dense_weights(layer, input_dims):
    # Quantize weights
    weights = layer.get_weights()[0]
    wq, delta = quantize_weights(layer.quantizer, weights)

    # retrieve input dimensions from Akida's layer
    inwidth, inheight, inchans = input_dims
    # Kernels in the fully connected are in the (HxWxC,N) format, more
    # specifically in each neuron data is laid out in the H,W,C format. In
    # Akida we expect a kernel in the (N,CxHxW,1,1), where the data in each
    # neuron is laid out in the W,H,C format as the input dimensions are
    # set.
    # So the operations done in order to obtain the akida fully connected
    # kernel are:
    # 1. reshape to H,W,C,N to split data across dimensions
    # 2. transpose dimensions to obtain C,H,W,N
    # 3. reshape to: 1,1,CxHxW,N
    #
    wq_akida = wq.reshape(inheight, inwidth, inchans, layer.units) \
        .transpose(2, 0, 1, 3) \
        .reshape(1, 1, inchans * inheight * inwidth, layer.units)
    # Multiply by delta, round and cast to int
    wq_akida = wq_akida * delta
    wq_akida = np.round(wq_akida).astype(np.int8)

    return wq_akida, delta


def _prepare_threshold_calculations(layer_map, layers_keras):
    """Retrieves required variables to calculate Akida fire threshold and
    step.

    This function returns a dictionary with variables required for threshold
    calculations of the given Akida layer map: bias in the Keras neural layer,
    parameters of the activation layer, batch norm parameters and global
    average pooling factor.

    Args:
        layer_map (LayerMapping): the given Akida layer map.
        layers_keras: the list of Keras layers (tf.keras.model.layers)
            corresponding to the given layer map.

    Returns:
        dict: required variables to calculate Akida fire threshold and step.
    """

    layer_neural = layers_keras[layer_map.index_neural]

    # Initialize variables to their default values
    threshold_shape = layer_neural.output_shape[-1]
    bias = np.zeros(threshold_shape)
    thresh_keras = np.zeros(threshold_shape)
    step_keras = np.ones(threshold_shape)
    gamma_bn = np.ones(threshold_shape)
    sigma_bn = np.ones(threshold_shape)
    beta_bn = np.zeros(threshold_shape)
    mu_bn = np.zeros(threshold_shape)
    global_avg_pool_factor = 1

    # Get bias if present
    if layer_neural.use_bias:
        if layer_map.layer_type == LayerType.SeparableConvolutional:
            bias = layer_neural.get_weights()[2]
        else:
            bias = layer_neural.get_weights()[1]

    # Get parameters from the discrete ReLU activation (threshold and step)
    if layer_map.index_activation:
        act_layer = layers_keras[layer_map.index_activation]
        thresh_keras = act_layer.threshold.numpy() * np.ones(threshold_shape)
        step_keras = (act_layer.step_width.numpy() * 2.**act_layer.bitwidth /
                      16) * np.ones(threshold_shape)

    # Get batchnorm parameters
    if layer_map.index_batchnorm:
        (gamma_bn, beta_bn, mu_bn,
         var_bn) = layers_keras[layer_map.index_batchnorm].get_weights()
        sigma_bn = np.sqrt(var_bn +
                           layers_keras[layer_map.index_batchnorm].epsilon)

    # Get global average pooling factor
    if layer_map.index_pool and isinstance(layers_keras[layer_map.index_pool],
                                           layers.GlobalAveragePooling2D):
        global_avg_pool_factor = np.prod(
            layers_keras[layer_map.index_pool].input_shape[1:3])

    return KerasLayerParams(bias, thresh_keras, step_keras, gamma_bn, sigma_bn,
                            beta_bn, mu_bn, global_avg_pool_factor)


def convert_layer_thresholds(layer_map, layers_keras, input_scaling,
                             weights_scale_factor, weights_ak):
    """Returns the fire thresholds and threshold steps for a given Akida layer.

    This function computes the fire thresholds and steps for a given Akida layer
    (defined by the layer map) from the corresponding Keras layers. Other
    arguments are required to compute these variables.

    Args:
        layer_map (LayerMapping): the given Akida layer map
        layers_keras: the list of Keras layers (tf.keras.model.layers)
            corresponding to the given layer map
        input_scaling (2-element tuple): the input scaling of the current layer.
        weights_scale_factor (float): the weights scale factor of the current
            Akida layer. This is given by the 'convert_layer_weights' function.
        weights_ak (np.ndarray): the Akida weights of the current Akida layer.

    Returns:
        tuple: the fire thresholds and steps.

    """

    input_scale, input_shift = input_scaling
    p = _prepare_threshold_calculations(layer_map, layers_keras)

    # Compute threshold fires and steps
    th_fire = (input_scale * weights_scale_factor *
               (p.sigma_bn / p.gamma_bn *
                (p.act_threshold - p.beta_bn) + p.mu_bn - p.bias))
    th_step = (input_scale * weights_scale_factor * p.sigma_bn / p.gamma_bn *
               p.act_step).astype(np.float32)

    # If there is an InputConv layer, update th_fire with input shift
    if layer_map.layer_type == LayerType.InputConvolutional:
        th_fire += np.sum(weights_ak, axis=(0, 1, 2)) * input_shift

    th_fire *= p.global_avg_pool_factor
    th_step *= p.global_avg_pool_factor

    th_fire = np.floor(th_fire).astype(np.int32)

    return th_fire, th_step


def quantize_weights(quantizer, w):
    """Returns quantized weights and delta as numpy arrays.

    Internally, it uses a tf.function that wraps calls to the quantizer in
    a graph, allowing the weights to be quantized eagerly.

    Args:
        quantizer (:obj:`WeightQuantizer`): the quantizer object.
        w (:obj:`np.ndarray`): the weights to quantize.

    Retruns:
        :obj:`np.ndarray`: the quantized weights `np.ndarray` and the scale
            factor scalar.

    """

    w_tf = tf.constant(w)
    wq = quantizer.quantize(w_tf)
    scale_factor = quantizer.scale_factor(w_tf)
    return wq.numpy(), scale_factor.numpy()



# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Utilities for API compatibility between CNN2SNN release versions"""

import os
import shutil
import h5py
from tensorflow.keras.models import model_from_json
from cnn2snn import load_quantized_model, cnn2snn_objects


def _upgrade_model_with_legacy_quantizers(src_filepath, dst_filepath):
    """Replaces legacy quantizers (cnn2snn<=1.8.9) with new ones
    (cnn2snn>=1.8.10).

    The model configuration is retrieved and modified to replace the legacy
    quantizers "WeightQuantizer" and "TrainableWeightQuantizer" with the new
    "StdWeightQuantizer" and "TrainableStdWeightQuantizer". A new model
    with the weights of the source model is then saved.

    HDF5 and TF saved models are supported.

    Args:
        src_filepath: model filepath with legacy quantizers
        dst_filepath: filepath to save the upgraded model
    """

    if not dst_filepath:
        raise ValueError("The destination model filepath must not be empty")

    def _replace_legacy_quantizers(model_config):
        model_config = model_config.replace('"WeightQuantizer"',
                                            '"StdWeightQuantizer"')
        model_config = model_config.replace('"TrainableWeightQuantizer"',
                                            '"TrainableStdWeightQuantizer"')
        return model_config

    if h5py.is_hdf5(src_filepath):

        # Copy and open the h5 model file
        shutil.copy(src_filepath, dst_filepath)
        f = h5py.File(dst_filepath, mode='r+')

        # Get model configuration
        model_config = f.attrs.get('model_config')
        if model_config is None:
            raise ValueError('No model found in config file.')
        model_config = model_config.decode('utf-8')

        # Detect if there are legacy quantizers
        if ('"WeightQuantizer"' in model_config or
                '"TrainableWeightQuantizer"' in model_config):

            # Replace legacy quantizers with their new names
            model_config = _replace_legacy_quantizers(model_config)

            # Update model config in the h5py.File
            f.attrs['model_config'] = model_config.encode('utf8')
            f.close()

            print(f"Model upgraded and saved to {dst_filepath}")
        else:
            os.remove(dst_filepath)
            print("No legacy quantizers found. No operation is done")

    else:  # If model file is in TF/PB format

        # Get model configuration
        model = load_quantized_model(src_filepath)
        model_config = model.to_json()

        # Detect if there are legacy quantizers
        if ('"WeightQuantizer"' in model_config or
                '"TrainableWeightQuantizer"' in model_config):

            # Replace legacy quantizers with their new names
            model_config = _replace_legacy_quantizers(model_config)

            # Save new model with source weights
            new_model = model_from_json(model_config,
                                        custom_objects=cnn2snn_objects)
            new_model.set_weights(model.get_weights())
            new_model.save(dst_filepath)

            print(f"Model upgraded and saved to {dst_filepath}")
        else:
            print("No legacy quantizers found. No operation is done")



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2019 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Conversion of a Keras/CNN2SNN model into an Akida model"""

import os
import tensorflow as tf
from . import utils
from .mapping_generator import generate_model_mapping, check_mapping_compatibility
from .model_generator import generate_model
from .weights_ops import set_weights_thresholds


def convert(model, file_path=None, input_scaling=None, input_is_image=True):
    """Converts a Keras quantized model to an Akida one.

    After quantizing a Keras model with :func:`cnn2snn.quantize`, it can be
    converted to an Akida model. By default, the conversion expects that the
    Akida model takes 8-bit images as inputs. ``input_scaling`` defines how the
    images have been rescaled to be fed into the Keras model (see note below).

    If inputs are spikes, you can set ``input_is_image=False``. In this case,
    Akida inputs are then expected to be integers between 0 and 15.

    Note:
        The relationship between Keras and Akida inputs is defined as::

            input_akida = input_scaling[0] * input_keras + input_scaling[1].

        If a :class:`tf.keras.layers.experimental.preprocessing.Rescaling`
        layer is present as first layer of the model, ``input_scaling`` must
        be None: the :class:`Rescaling` parameters will be used to compute the
        input scaling.

    Examples:

        >>> # Convert a quantized Keras model with Keras inputs as images
        >>> # rescaled between -1 and 1
        >>> inputs_akida = images.astype('uint8')
        >>> inputs_keras = (images.astype('float32') - 128) / 128
        >>> model_akida = cnn2snn.convert(model_keras, input_scaling=(128, 128))
        >>> model_akida.evaluate(inputs_akida)

        >>> # Convert a quantized Keras model with Keras inputs as spikes and
        >>> # input scaling of (2.5, 0). Akida spikes must be integers between
        >>> # 0 and 15
        >>> inputs_akida = spikes.astype('uint8')
        >>> inputs_keras = spikes.astype('float32') / 2.5
        >>> model_akida = cnn2snn.convert(model_keras, input_scaling=(2.5, 0))
        >>> model_akida.evaluate(inputs_akida)

        >>> # Convert and directly save the Akida model to fbz file.
        >>> cnn2snn.convert(model_keras, 'model_akida.fbz')

    Args:
        model (:obj:`tf.keras.Model`): a tf.keras model
        file_path (str, optional): destination for the akida model.
            (Default value = None)
        input_scaling (2 elements tuple, optional): value of the input scaling.
            (Default value = None)
        input_is_image (bool, optional): True if input is an image (3-D 8-bit
            input with 1 or 3 channels) followed by QuantizedConv2D. Akida model
            input will be InputConvolutional. If False, Akida model input will
            be InputData. (Default value = True)

    Returns:
        :obj:`akida.Model`: an Akida model.

    Raises:
        ValueError: If ``input_scaling[0]`` is null or negative.
        ValueError: If a :class:`Rescaling` layer is present and
            ``input_scaling`` is not None.
        SystemError: If Tensorflow is not run in eager mode.
    """

    if not tf.executing_eagerly():
        raise SystemError("Tensorflow eager execution is disabled. "
                          "It is required to convert Keras weights to Akida.")

    # Merge separable convolution
    model_sep = utils.merge_separable_conv(model)

    # Generate model mapping
    model_map = generate_model_mapping(model_sep, input_is_image)

    # Check Keras Rescaling layer to replace the input_scaling
    rescaling_input_scaling = _get_rescaling_layer_params(model_sep)
    if rescaling_input_scaling is not None and input_scaling is not None:
        raise ValueError("If a Rescaling layer is present in the model, "
                         "'input_scaling' argument must be None. Receives "
                         f"{input_scaling}.")

    input_scaling = rescaling_input_scaling or input_scaling or (1, 0)

    if input_scaling[0] <= 0:
        raise ValueError("The scale factor 'input_scaling[0]' must be strictly"
                         f" positive. Receives: input_scaling={input_scaling}")

    # Check compatibility of the model map
    check_mapping_compatibility(model_map)

    # Generate Akida model
    ak_inst = generate_model(model_map, input_scaling)

    # Convert weights
    set_weights_thresholds(model_map, ak_inst, input_scaling)

    # Save model if file_path is given
    if file_path:
        # Create directories
        dir_name, base_name = os.path.split(file_path)
        if base_name:
            file_root, file_ext = os.path.splitext(base_name)
            if not file_ext:
                file_ext = '.fbz'
        else:
            file_root = model.name
            file_ext = '.fbz'

        if dir_name and not os.path.exists(dir_name):
            os.makedirs(dir_name)

        save_path = os.path.join(dir_name, file_root + file_ext)
        ak_inst.save(save_path)

    return ak_inst


def _get_rescaling_layer_params(model):
    """Computes the new input scaling retrieved from the Keras
    `Rescaling` layer.

    Keras Rescaling layer works as:

     input_k = scale * input_ak + offset

    CNN2SNN input scaling works as:

     input_ak = input_scaling[0] * input_k + input_scaling[1]

    Equivalence leads to:

     input_scaling[0] = 1 / scale
     input_scaling[1] = -offset / scale

    Args:
        model (:obj:`tf.keras.Model`): a tf.keras model.

    Returns:
        tuple: the new input scaling from the Rescaling layer or None if
            no Rescaling layer is at the beginning of the model.
    """

    Rescaling = tf.keras.layers.experimental.preprocessing.Rescaling
    for layer in model.layers[:2]:
        if isinstance(layer, Rescaling):
            return (1 / layer.scale, -layer.offset / layer.scale)
    return None
